/*	
	Jeff Braun (March 2001)
	email1: yobkadon@hotmail.com
	email2: brauns@und-alum.org
*/
//#define DEBUG 1

#include <Rect.h>
#include <ScrollBar.h>
#include <View.h>

#include <Beep.h>
#include <Debug.h>
#include <Region.h>
#include <MessageRunner.h>
#include <ScrollView.h>
#include <Font.h>
#include <Invoker.h>
#include <List.h>
#include <Message.h>
#include <Point.h>
#include <string.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>


template<class T>
class FastListView : public BView, public BInvoker
{
	private:
		struct Node
		{
			int32 index;
			T * pObject;
			Node * prev;
			Node * next;
		};

	public:
	/***************************************************************
	***************************************************************/
	FastListView
		(
		BRect frame
	,	const char * name
	,	int32 length
	,	uint32 resizeMask = B_FOLLOW_LEFT | B_FOLLOW_TOP
	,	uint32 flags = B_WILL_DRAW | B_FRAME_EVENTS | B_NAVIGABLE
		)
	:	BView (frame, name, resizeMask, flags)
	,	m_Head (NULL)
	,	m_ArraySize (length)
	,	m_ItemCount (0l)
	,	m_array (NULL)
	,	m_pVBar (NULL)
	,	m_pHBar (NULL)
	,	m_runner (NULL)
	,	m_th (0l)
	,	m_thTotal (0l)
	,	m_start (1l)
	,	m_stop (1l)
	,	m_lastpos (1l)
	,	m_ItemsPerPage (1l)
	,	m_Interval (0.0f)
	,	m_SelectedItem (NULL)
	,	m_SelectedIndex (-1)
	,	m_TargetByScrollView (false)
	,	m_OwnerScrollView (NULL)
	,	m_NewItemsToAddList ()
	,	m_Tracking (false)
	,	m_MouseAbove (false)
	,	m_IncBy (0l)
	{
		PRINT (("FastListView::FastListView\n"));
		if (m_ArraySize < 2)
		{
			m_ArraySize = 2;
		}
		m_array = new Node[m_ArraySize];

		if (m_array)
		{
			PRINT (("  array created\n"));
			m_array[0].index = 0;
			m_array[0].pObject = NULL;
			m_array[0].prev = &(m_array[m_ArraySize - 1]); //this creates a circular loop!
			m_array[0].next = &(m_array[1]); //this creates a circular loop!
			int32 i (1);

			for (; i < m_ArraySize - 1; i++)
			{
				m_array[i].index = i;
				m_array[i].pObject = NULL;
				m_array[i].prev = &(m_array[i - 1]);
				m_array[i].next = &(m_array[i + 1]);
			}

			PRINT (("  finish last node\n"));
			m_array[i].index = i;
			m_array[i].pObject = NULL;
			m_array[i].prev = &(m_array[i - 1]); //this creates a circular loop!
			m_array[i].next = &(m_array[0]); //this creates a circular loop!

			PRINT (("  set Head\n"));
			m_Head = &(m_array[m_ArraySize - 1]);
			PRINT (("  m_Head.index = %i\n", m_Head->index));
		}

		BFont font;
		font_height fh;
		GetFont (&font);
		font.GetHeight (&fh);
		m_th = (int32) (fh.ascent + fh.leading);
		m_thTotal = m_th + (int32) fh.descent;
		m_thTotal += 4; //some buffer space
		PRINT (("  m_th = %i\tm_thTotal = %i\n", m_th, m_thTotal));
	}


	/***************************************************************
	***************************************************************/
	virtual
	~FastListView
		(
		void
		)
	{
		PRINT (("FastListView::~FastListView\n"));
		if (m_array)
		{
			T * pObject (NULL);
			for (int32 i (0); i < m_ArraySize; i++)
			{
				pObject = m_array[i].pObject;
				if (pObject)
				{
					delete pObject;
					m_array[i].pObject = NULL;
				}
			}
			delete [] m_array;
		}
	}


	public:
	/***************************************************************
	***************************************************************/
	virtual void
	MessageReceived
		(
		BMessage * msg
		)
	{
		switch (msg->what)
		{
			case 'scrl':
				if (m_MouseAbove)
				{
					PRINT (("scroll up\n"));
					Up (m_IncBy);
				}
				else
				{
					PRINT (("scroll down\n"));
					Down (m_IncBy);
				}
				break;
			default:
				BView::MessageReceived (msg);
		}
	}


	public:
	/***************************************************************
	***************************************************************/
	virtual void
	Draw
		(
		BRect updateRect
		)
	{
		m_start = (int32) (updateRect.top / m_thTotal) - 1;
		if (m_start < 0)
		{
			m_start = 0;
		}

		if (m_start >= m_ArraySize)
		{
			PRINT (("Draw quickexit\n"));
			return;
		}

		m_stop = (int32) (updateRect.bottom / m_thTotal) + 1 ;

		if (m_stop > m_ArraySize)
		{
			m_stop = m_ArraySize;
		}
	
		PRINT (("\nDraw\n  IPP = %i\tStart = %i\tStop = %i\n", m_ItemsPerPage, m_start, m_stop));
		PRINT (("  rect: l=%0.f\tt=%0.f\tr=%0.f\tb=%0.f\n", updateRect.left,updateRect.top,updateRect.right, updateRect.bottom));
	
		PRINT (("  Head.index = %li\n", m_Head->index));
		Node * cur (m_Head);
		int32 index (cur->index);

		if (index - m_start < 0)
		{
			PRINT (("  deal with index - m_start < 0:\tindex = "));
			index = m_ArraySize + index - m_start;
		}
		else
		{
			index -= m_start;
			PRINT (("  set index -= m_start:\tindex = "));
		}

		PRINT (("  %i\n", index));
		cur = &(m_array[index]);

		//BRect rc (Frame ());
		BRect rc (Bounds ());
		PRINT ((" Frame.left = %f\n", rc.left));
		rc.left = 0.0f;
		
		BRegion regn;

		for (int32 i (m_start); i < m_stop && cur->pObject; i++)
		{
			//PRINT (("  Node.Index = %i\n", cur->index));
			rc.top = i * m_thTotal;
			rc.bottom = rc.top + m_thTotal - 1.0f;
			regn.Set (rc);
			BView::ConstrainClippingRegion (&regn);
			BRect rcOther (0,0,0,0);
			cur->pObject->DrawItem (this, rc, &rcOther, 9, false);
			cur = cur->prev;
		}

		BView::ConstrainClippingRegion (NULL);
		Sync ();
	}


	public:
	/***************************************************************
	***************************************************************/
	virtual void
	AttachedToWindow
		(
		void
		)
	{
		PRINT (("\nAttachedToWindow\n"));
		m_ItemsPerPage = (int32) Bounds().Height () / m_thTotal;
		++m_ItemsPerPage;
		m_pHBar = ((BScrollView *) Parent ())->ScrollBar (B_HORIZONTAL);
		m_pVBar = ((BScrollView *) Parent ())->ScrollBar (B_VERTICAL);
	
		m_pVBar->SetRange (0, m_ItemCount * m_thTotal);
		PRINT (("  SetRange to 0, %i.\n", m_ItemCount * m_thTotal));
	
		m_pVBar->SetSteps (m_thTotal, 10 * m_thTotal);
		PRINT (("  SetSteps to %i, %i\n", m_thTotal, 10 * m_thTotal));
	}


	public:
	/***************************************************************
	***************************************************************/
	virtual void
	MakeFocus
		(
		bool focusState = true
		)
	{
		BView::MakeFocus (focusState);
		if (m_TargetByScrollView)
		{
			m_OwnerScrollView->SetBorderHighlighted (focusState);
		}
	}
	

	public:
	/***************************************************************
	***************************************************************/
	virtual void
	WindowActivated
		(
		bool state
		)
	{
		if (m_TargetByScrollView && IsFocus ())
		{
			m_OwnerScrollView->SetBorderHighlighted (state);
		}
	}


	public:
	/***************************************************************
	***************************************************************/
	virtual void
	TargetedByScrollView
		(
		BScrollView * scroller
		)
	{
		PRINT (("\nTargetedByScrollView\n"));
		BView::TargetedByScrollView (scroller);
		m_TargetByScrollView = true;
		m_OwnerScrollView = scroller;
	}


	public:
	/***************************************************************
	***************************************************************/
	virtual void
	ScrollTo
		(
		BPoint where
		)
	{
		BView::ScrollTo (where);
		PRINT (("\nScrollTo:: where: x=%f,   y=%f\n", where.x, where.y));
		BRect rc (Frame ());
		PRINT (("  new frame: l=%f\tt=%f\tr=%f\tb=%f\n",rc.left,rc.top,rc.right,rc.bottom));
		rc = Bounds ();
		PRINT (("  new bounds: l=%f\tt=%f\tr=%f\tb=%f\n",rc.left,rc.top,rc.right,rc.bottom));
	}

	
	public:
	/***************************************************************
	***************************************************************/
	virtual void
	ScrollToSelection
		(
		void
		)
	{
		PRINT (("\nScrollToSelection\n"));
		int32 realindex (0);
		BRect b (Bounds ());
#ifdef DEBUG
		BRect f (Frame ());
#endif

		PRINT (("  bounds: l=%f\tt=%f\tr=%f\tb=%f\n",b.left,b.top,b.right,b.bottom));
		PRINT (("  frame: l=%f\tt=%f\tr=%f\tb=%f\n",f.left,f.top,f.right,f.bottom));
		PRINT (("  SelIndex = %i\tHead.index = %i\n", m_SelectedIndex, m_Head->index));
		
		if (m_SelectedIndex <= m_Head->index)
		{
			realindex = m_Head->index - m_SelectedIndex;
		}
		else
		{
			realindex = m_ArraySize - m_SelectedIndex + m_Head->index;
		}

		PRINT (("  realindex = %i\tIPP = %i\n", realindex, m_ItemsPerPage));
		
		if (realindex * m_thTotal < b.top) //above the window, scroll up
		{
			m_pVBar->SetValue (realindex * m_thTotal);
			PRINT (("  ScrollUp setting value to %f\n", realindex * m_thTotal));
		}
		else if ((realindex+1) * m_thTotal > b.bottom) //below the window, scroll down
		{
			float diff (((realindex+1) * m_thTotal) - b.bottom);
			diff = m_pVBar->Value () + diff - 1.0f;
			m_pVBar->SetValue (diff);
			PRINT (("  ScrollDown setting value to %f\n", diff));
		}
		else //we're currently visible?!
		{
			PRINT (("  already visible!\n\n"));
		}
	}


	public:
	/***************************************************************
	***************************************************************/
	void
	ResizeBy
		(
		float dh
	,	float dv
		)
	{
		PRINT (("ResizeBy called\n"));
		BView::ResizeBy (dh, dv);
		m_ItemsPerPage = (int32) Bounds().Height () / m_thTotal;
		++m_ItemsPerPage;
		if (m_pVBar)
		{
			float propH = new_height / (m_ArraySize * m_thTotal);
			PRINT (("  SetProportion %f\n", propH));
			m_pVBar->SetProportion (propH);
			float rangeH = (m_ArraySize * m_thTotal) - new_height - 1.0f;
			if(rangeH < 0) rangeH = 0;
			PRINT (("  SetRange 0, %f\n", rangeH));
			m_pVBar->SetRange (0, rangeH);
			PRINT (("  SetSteps %f, %f\n", new_height / 8.0, new_height / 2.0));
			m_pVBar->SetSteps((float) m_thTotal, m_thTotal * 10.0f);
		}
	}


	public:
	/***************************************************************
	***************************************************************/
	void
	ResizeTo
		(
		float new_width
	,	float new_height
		)
	{
		PRINT (("ResizeTo called\n"));
		BView::ResizeTo (new_width, new_height);
		m_ItemsPerPage = (int32) Bounds().Height () / m_thTotal;
		++m_ItemsPerPage;

		if (m_pVBar)
		{
			float propH = new_height / (m_ArraySize * m_thTotal);
			PRINT (("  SetProportion %f\n", propH));
			m_pVBar->SetProportion (propH);
			float rangeH = (m_ArraySize * m_thTotal) - new_height - 1.0f;
			if(rangeH < 0) rangeH = 0;
			PRINT (("  SetRange 0, %f\n", rangeH));
			m_pVBar->SetRange (0, rangeH);
			PRINT (("  SetSteps %f, %f\n", new_height / 8.0, new_height / 2.0));
			m_pVBar->SetSteps((float) m_thTotal, m_thTotal * 10.0f);
		}
	}


	public:
	/***************************************************************
	***************************************************************/
	virtual void
	FrameResized
		(
		float new_width
	,	float new_height
		)
	{
		PRINT (("FrameResized: w=%f\th=%f\n", new_width, new_height));
		BView::FrameResized (new_width, new_height);
		m_ItemsPerPage = (int32) Bounds().Height () / m_thTotal;
		++m_ItemsPerPage;

		if (m_pVBar)
		{
			float propH = new_height / (m_ItemCount * m_thTotal);
			PRINT (("  SetProportion %f\n", propH));
			m_pVBar->SetProportion (propH);
			float rangeH = (m_ItemCount * m_thTotal) - new_height - 1.0f;
			if(rangeH < 0) rangeH = 0;
			//m_pVBar->SetRange (0, m_ArraySize * m_thTotal - m_ItemsPerPage);
			PRINT (("  SetRange 0, %f\n", rangeH));
			m_pVBar->SetRange (0, rangeH);
			PRINT (("  SetSteps %f, %f\n", new_height / 8.0, new_height / 2.0));
			m_pVBar->SetSteps((float) m_thTotal, m_thTotal * 10.0f);
		}
	}

	public:
	/***************************************************************
	***************************************************************/
	virtual void
	AddItem
		(
		T * NewObject
		)
	{
		m_NewItemsToAddList.AddItem ((void *) NewObject);
	}


	private:
	/***************************************************************
	***************************************************************/
	void
	InternalAdd
		(
		void
		)
	{
		PRINT (("InternalAdd()\n"));
		int32 count (m_NewItemsToAddList.CountItems ());
		int32 inc (0);

		do
		{
			PRINT (("  current Head.index = %i\n", m_Head->index));
			m_Head = m_Head->next;
			if (m_Head->index == m_SelectedIndex)
			{
				//then we've just wiped out the currently selected item so clear it
				m_SelectedIndex = -1;
				m_SelectedItem = NULL;
				PRINT (("  RESET: the current selection slid off the bottom\n"));
			}

			int32 index (m_Head->index);
			T * pObject (m_array[index].pObject);

			if (pObject)
			{
				PRINT (("  there was data in the object, delete it...\n"));
				delete pObject;
			}

			if (++m_ItemCount > m_ArraySize)
			{
				m_ItemCount = m_ArraySize;
			}
			m_array[index].pObject = reinterpret_cast<T *> (m_NewItemsToAddList.ItemAt (inc));
			PRINT (("  new Head.index = %i\n", m_Head->index));
		}
		while (++inc < count);
	}


	public:
	/***************************************************************
	***************************************************************/
	void
	DisableForAdd
		(
		bool IsDisabled
		)
	{
		int32 OverFlow (0);

		if (IsDisabled)
		{
			m_NewItemsToAddList.MakeEmpty ();
		}
		else
		{
			if (m_NewItemsToAddList.CountItems () > 0)
			{
				int32 ItemsToBeAdded (m_NewItemsToAddList.CountItems ());
				if ((ItemsToBeAdded + m_ItemCount) > m_ArraySize)
				{
					OverFlow = (ItemsToBeAdded + m_ItemCount) - m_ArraySize;
				}
				BRect source (Bounds ());
				if ((ItemsToBeAdded * m_thTotal) < source.Height ())
				{
					if (OverFlow > 0)
					{
						if ((m_ItemCount * m_thTotal) < source.bottom)
						{
							source.bottom = (m_ItemCount - OverFlow) * m_thTotal - 1.0f;
						}
						//source.bottom -= OverFlow * m_thTotal;
					}
					if (source.bottom > 0)
					{
						PRINT (("do CopyBits (0, %i)\n", ItemsToBeAdded * m_thTotal));
						BRect destination (source.OffsetByCopy (0.0f, ItemsToBeAdded * m_thTotal));
						Window()->UpdateIfNeeded ();
						Sync ();
						CopyBits (source, destination);
					}
					BRect invalid (source);
					invalid.bottom = source.top + (ItemsToBeAdded * m_thTotal);
					Invalidate (invalid);
				}
				else
				{
					PRINT (("do a plain Invalidate()\n"));
					Invalidate ();
				}
				
				InternalAdd (); //add the items now!

				float height (Bounds().Height ());
				float propH = height / (m_ItemCount * m_thTotal);
				//PRINT (("  SetProportion %f\n", propH));
				m_pVBar->SetProportion (propH);
				float rangeH = (m_ItemCount * m_thTotal) - height - 1.0f;
				if(rangeH < 0) rangeH = 0;
				//PRINT (("  SetRange 0, %f\n", rangeH));
				m_pVBar->SetRange (0, rangeH);

				Window()->UpdateIfNeeded ();

				m_NewItemsToAddList.MakeEmpty ();
				Sync ();
			}
		}
	}


	public:
	/***************************************************************
	***************************************************************/
	virtual void
	Up
		(
		uint32 by
		)
	{
		int32 index (0);
		if (m_SelectedIndex < 0) //then nothing is selected currently
		{
			index = m_Head->index;
			Select (OffsetOf (index));
		}
		else
		{
			if (m_Head != &(m_array[m_SelectedIndex]))
			{
				Node * pNode (&(m_array[m_SelectedIndex]));
				while (by-- > 0 && m_Head != pNode)
				{
					pNode = pNode->next;
				}

				index = pNode->index;
				Select (OffsetOf (index));
			}
			//otherwise we're at the top and can't go any higher
		}

		ScrollToSelection ();
	}


	public:
	/***************************************************************
	***************************************************************/
	virtual void
	Down
		(
		uint32 by
		)
	{
		int32 index (0);
		if (m_SelectedIndex < 0) //then nothing is selected currently
		{
			index = m_Head->index;
			Select (OffsetOf (index));
		}
		else
		{
			Node * Tail (GetTail ());
			if (Tail != &m_array[m_SelectedIndex])
			{
				Node * pNode (&(m_array[m_SelectedIndex]));
				while (by-- > 0 && Tail != pNode)
				{
					pNode = pNode->prev;
				}
				index = pNode->index;
				Select (OffsetOf (index));
			}
			//otherwise we're at the top and can't go any higher
		}

		ScrollToSelection ();
	}
	
	
	private:
	/***************************************************************
	***************************************************************/
	Node *
	GetTail
		(
		void
		)
	{
		int32 HeadIndex (m_Head->index);
		int32 TailIndex (0);
		if ((HeadIndex - m_ItemCount + 1) < 0)
		{
			TailIndex = m_ArraySize - (m_ItemCount - HeadIndex) + 1;
		}
		else
		{
			TailIndex = (HeadIndex - m_ItemCount) + 1;
		}
		return (reinterpret_cast<Node *> (&(m_array[TailIndex])));
	}

	
	public:
	/***************************************************************
	***************************************************************/
	virtual void
	MouseDown
		(
		BPoint where
		)
	{
		MakeFocus (true);
		Select (IndexOf (where));
		SetEventMask (B_POINTER_EVENTS, B_NO_POINTER_HISTORY);
		m_Tracking = true;
	}
	

	public:
	/***************************************************************
	***************************************************************/
	virtual void
	MouseMoved
		(
		BPoint pt
	,	uint32 transit
	,	const BMessage * msg
		)
	{
		PRINT (("MouseMoved: x=%f\ty=%f\n", pt.x, pt.y));
		float Y (ConvertToScreen (pt).y);
		float diff (0.0f);
		float interval (0.0f);
		BRect rc (ConvertToScreen (Bounds ()));

		if (m_Tracking)
		{
			if (Y < rc.top)
			{
				m_MouseAbove = true;
				diff = rc.top - Y;
				m_IncBy = ((int32) diff / m_thTotal) + 1;
				interval = 50000;

				PRINT (("  interval = %f\n", interval));
				if (NULL == m_runner)
				{
					BMessage msg ('scrl');
					m_runner = reinterpret_cast<BMessageRunner *> (new BMessageRunner (BMessenger (this), &msg, interval));
				}
				else if(interval != m_Interval)
				{
					m_Interval = interval;
					delete m_runner;
					BMessage msg ('scrl');
					m_runner = reinterpret_cast<BMessageRunner *> (new BMessageRunner (BMessenger (this), &msg, interval));
				}
				PRINT (("  ABOVE\n"));
			}
			else if (Y > rc.bottom)
			{
				PRINT (("  below\n"));
				m_MouseAbove = false;
				diff = Y - rc.bottom;
				m_IncBy = ((int32) diff / m_thTotal) + 1;
				interval = 50000;

				PRINT (("  interval = %f\n", interval));
				if (NULL == m_runner)
				{
					BMessage msg ('scrl');
					m_runner = reinterpret_cast<BMessageRunner *> (new BMessageRunner (BMessenger (this), &msg, interval));
				}
				else if(interval != m_Interval)
				{
					m_Interval = interval;
					delete m_runner;
					BMessage msg ('scrl');
					m_runner = reinterpret_cast<BMessageRunner *> (new BMessageRunner (BMessenger (this), &msg, interval));
				}
			}
			else
			{
				PRINT (("  Middle\n"));
				if (NULL != m_runner)
				{
					PRINT (("\n\n  delete RUNNER!!!\n\n\n"));
					delete m_runner;
					m_runner = NULL;
				}
				Select (IndexOf (pt));
			}
		}
	}


	public:
	/***************************************************************
	***************************************************************/
	virtual void
	MouseUp
		(
		BPoint pt
		)
	{
		PRINT (("\n\nMouseUp: x=%f\ty=%f\n\n", pt.x, pt.y));
		if (m_Tracking)
		{
			SetEventMask (0, 0);
			m_Tracking = false;
			if (NULL != m_runner)
			{
				delete m_runner;
				m_runner = NULL;
			}
		}
	}
	

	public:
	/***************************************************************
	***************************************************************/
	virtual void
	KeyDown
		(
		const char * bytes
	,	int32 numBytes
		)
	{
		PRINT (("KeyDown: SelIndex = %i\n", m_SelectedIndex));
		float bigstep (0.0f), smallstep (0.0f);
		float val (0.0f);

		switch (*bytes)
		{
			case B_UP_ARROW:
				Up (1);
				break;
			case B_DOWN_ARROW:
				Down (1);
				break;
			case B_PAGE_UP:
				Up (m_ItemsPerPage - 1);
				break;
			case B_PAGE_DOWN:
				Down (m_ItemsPerPage - 1);
				break;
			case B_RIGHT_ARROW:
				m_pHBar->GetSteps (&smallstep, &bigstep);
				val = m_pHBar->Value ();
				m_pHBar->SetValue (val + bigstep);
				break;
			case B_LEFT_ARROW:
				m_pHBar->GetSteps (&smallstep, &bigstep);
				val = m_pHBar->Value ();
				m_pHBar->SetValue (val - bigstep);
				break;
			default:
				BView::KeyDown (bytes, numBytes);
				break;
		}
	}
	

	public:
	/***************************************************************
	***************************************************************/
	int32
	IndexOf (BPoint point) const
	{
		int32 selindex ((int32) point.y / m_thTotal);
		PRINT (("IndexOf: point %f, %f\titem index would be %i\n", point.x, point.y, selindex));
		if ((selindex > (m_ItemCount - 1) || (selindex < 0)))
		{
			return -1;
		}
		return selindex;
	}
	
	
	public:
	/***************************************************************
	***************************************************************/
	status_t
	Select (int32 selindex)
	{
		PRINT (("Select (%li)\n", selindex));
		if ((selindex < 0) || (selindex >= m_ItemCount))
		{
			PRINT (("  Bad index.\n"));
			return B_BAD_INDEX;
		}
		if (selindex == OffsetOf (m_SelectedIndex))
		{
			PRINT (("  Item already selected, ignore.\n"));
			return B_OK;
		}
		T * olditem (NULL);
		if (m_SelectedIndex >= 0)
		{
			olditem = m_array[m_SelectedIndex].pObject;
			if (NULL != olditem)
			{
				PRINT (("  deselect previous item.\n"));
				olditem->Deselect ();
				BRect rc (Bounds ());
				BRegion regn;
				rc.top = OffsetOf (m_SelectedIndex) * m_thTotal;
				rc.bottom = rc.top + m_thTotal - 1.0f;
				regn.Set (rc);
				BView::ConstrainClippingRegion (&regn);
				BRect rcOther (0,0,0,0);
				olditem->DrawItem (this, rc, &rcOther, 9, true);
			}
		}

		int32 index (m_Head->index);
		if (index - selindex < 0)
		{
			index = m_ArraySize + index - selindex;
		}
		else
		{
			index -= selindex;
		}

		m_SelectedItem = m_array[index].pObject;
		m_SelectedIndex = m_array[index].index;
		PRINT (("  SelIndex now = %i\n", m_SelectedIndex));
		if (m_SelectedItem)
		{
			PRINT (("  select the new item\n"));
			m_SelectedItem->Select ();
			BRect rc (Bounds ());
			BRegion regn;
			rc.top = OffsetOf (m_SelectedIndex) * m_thTotal;
			rc.bottom = rc.top + m_thTotal - 1.0f;
			regn.Set (rc);
			BView::ConstrainClippingRegion (&regn);
			BRect rcOther (0,0,0,0);
			m_SelectedItem->DrawItem (this, rc, &rcOther, 9, true);
		}

		BView::ConstrainClippingRegion (NULL);
		Sync ();

		return B_OK;
	}
	
	
	private:
	/***************************************************************
	***************************************************************/
	int32
	OffsetOf
		(
		int32 starti
		)
	{
		int32 headi (m_Head->index);
		int32 offset (0);

		if (starti <= headi)
		{
			offset = headi - starti;
		}
		else
		{
			offset = headi + (m_ArraySize - starti);
		}
		return offset;
	}


	private:
		FastListView (const FastListView &);
		FastListView & operator= (FastListView &);


	private:
		Node *			m_Head;
		int32			m_ArraySize;
		int32			m_ItemCount;
		Node *			m_array;
		BScrollBar *	m_pVBar;
		BScrollBar *	m_pHBar;
		BMessageRunner * m_runner;
		int32			m_th;
		int32			m_thTotal;
		int32			m_start;
		int32			m_stop;
		int32			m_lastpos;
		int32			m_ItemsPerPage;
		float			m_Interval;
		T *				m_SelectedItem;
		int32			m_SelectedIndex;
		bool			m_TargetByScrollView;
		BScrollView *	m_OwnerScrollView;
		BList			m_NewItemsToAddList;
		bool			m_Tracking;
		bool			m_MouseAbove;
		int32			m_IncBy;
};




